@@ -5,8 +5,9 @@ from pykusto.expressions import AnyTypeColumn, NumberType, NumberExpression, Tim
     DatetimeExpression, TimespanExpression, ArrayType, DynamicType, DatetimeType, BaseExpression, BooleanType, \
     ExpressionType, AggregationExpression, StringType, StringExpression, BooleanExpression, \
     NumberAggregationExpression, MappingAggregationExpression, ArrayAggregationExpression, to_kql, DynamicExpression, \
-    ArrayExpression, ColumnToType, BaseColumn
+    ArrayExpression, ColumnToType, BaseColumn, AnyExpression, AnyAggregationExpression, MappingExpression
 from pykusto.kql_converters import KQL
+from pykusto.logger import logger
 from pykusto.type_utils import plain_expression, get_base_types, KustoType
 
 
@@ -106,11 +107,11 @@ def bin_auto(expr: Union[NumberType, DatetimeType, TimespanType]) -> BaseExpress
 # def binary_xor(self): return
 
 
-def case(predicate: BooleanType, val: ExpressionType, *args: Union[BooleanType, ExpressionType]) -> BaseExpression:
+def case(predicate: BooleanType, val: ExpressionType, *args: Union[BooleanType, ExpressionType]) -> AnyExpression:
     res = 'case({}, {}, {})'.format(
         to_kql(predicate), to_kql(val), ', '.join([to_kql(arg) for arg in args])
     )
-    return AggregationExpression(KQL(res))
+    return AnyExpression(KQL(res))
 
 
 def ceiling(expr: NumberType) -> NumberExpression:
@@ -271,11 +272,17 @@ def iff(predicate: BooleanType, if_true: ExpressionType, if_false: ExpressionTyp
     common_types = other_types & return_types
     if len(common_types) == 0:
         # If there is not at least one common type, then certainly the arguments are not of the same type
-        raise TypeError("The second and third arguments must be of the same type, but they are: {} and {}".format(
-            ", ".join(t.primary_name for t in return_types),
-            ", ".join(t.primary_name for t in other_types)
-        ))
-    return plain_expression.registry[next(iter(common_types))](
+        logger.warning(
+            "The second and third arguments must be of the same type, but they are: {} and {}. "
+            "If this is a mistake, please report it at https://github.com/Azure/pykusto/issues".format(
+                ", ".join(sorted(t.primary_name for t in return_types)),
+                ", ".join(sorted(t.primary_name for t in other_types))
+            )
+        )
+        expression_type = AnyExpression
+    else:
+        expression_type = plain_expression.registry[next(iter(common_types))]
+    return expression_type(
         KQL('iff({}, {}, {})'.format(to_kql(predicate), to_kql(if_true), to_kql(if_false)))
     )
 
@@ -362,10 +369,10 @@ def make_datetime(year: NumberType,
     return DatetimeExpression(KQL(res))
 
 
-def make_string(): raise NotImplemented  # TODO
+def make_string() -> StringExpression: raise NotImplemented  # TODO
 
 
-def make_timespan(): raise NotImplemented  # TODO
+def make_timespan() -> TimespanExpression: raise NotImplemented  # TODO
 
 
 # def max_of(self): return
@@ -374,10 +381,10 @@ def make_timespan(): raise NotImplemented  # TODO
 # def min_of(self): return
 
 
-def monthofyear(): raise NotImplemented  # TODO
+def monthofyear() -> NumberExpression: raise NotImplemented  # TODO
 
 
-def new_guid(): raise NotImplemented  # TODO
+def new_guid() -> AnyExpression: raise NotImplemented  # TODO
 
 
 def now(offset: TimespanType = None) -> StringExpression:
@@ -386,16 +393,22 @@ def now(offset: TimespanType = None) -> StringExpression:
     return StringExpression(KQL('now()'))
 
 
-def pack(): raise NotImplemented  # TODO
+def pack(**kwargs: ExpressionType) -> MappingExpression:
+    return MappingExpression(KQL('pack({})'.format(
+        ', '.join('"{}", {}'.format(k, to_kql(v)) for k, v in kwargs.items())
+    )))
 
 
-def pack_all(): raise NotImplemented  # TODO
+def pack_all() -> MappingExpression: raise NotImplemented  # TODO
 
 
-def pack_array(): raise NotImplemented  # TODO
+def pack_array(*elements: ExpressionType) -> 'ArrayExpression':
+    return ArrayExpression(KQL('pack_array({})'.format(
+        ', '.join(to_kql(e) for e in elements)
+    )))
 
 
-def pack_dictionary(): raise NotImplemented  # TODO
+def pack_dictionary() -> MappingExpression: raise NotImplemented  # TODO
 
 
 #
@@ -428,10 +441,10 @@ def parse_json(expr: Union[StringType, DynamicType]) -> DynamicExpression:
 # def parse_xml(self): return
 
 
-def percentile_tdigest(): raise NotImplemented  # TODO
+def percentile_tdigest() -> AnyExpression: raise NotImplemented  # TODO
 
 
-def percentrank_tdigest(): raise NotImplemented  # TODO
+def percentrank_tdigest() -> AnyExpression: raise NotImplemented  # TODO
 
 
 def pow(expr1: NumberType, expr2: NumberType) -> NumberExpression:
@@ -673,15 +686,16 @@ def todouble(expr: NumberType) -> NumberExpression:
     return NumberExpression(KQL("todouble({})".format(to_kql(expr))))
 
 
-def todynamic(): raise NotImplemented  # TODO
+def todynamic() -> DynamicExpression: raise NotImplemented  # TODO
 
 
-def toguid(): raise NotImplemented  # TODO
+def toguid() -> AnyExpression: raise NotImplemented  # TODO
 
 
 def tohex(expr1: NumberType, expr2: NumberType = None) -> StringExpression:
     return StringExpression(KQL(('tohex({})' if expr2 is None else 'tohex({}, {})').format(to_kql(expr1), to_kql(expr2))))
 
+
 def toint(expr: NumberType) -> NumberExpression:
     return NumberExpression(KQL("toint({})".format(to_kql(expr))))
 
@@ -702,12 +716,13 @@ def tostring(expr: ExpressionType):
     return expr.to_string()
 
 
-def totimespan(): raise NotImplemented  # TODO
+def totimespan() -> TimespanExpression: raise NotImplemented  # TODO
 
 
 def toupper(expr: StringType) -> StringExpression:
     return expr.upper()
 
+
 # def to_utf8(self): return
 #
 #
@@ -717,7 +732,7 @@ def toupper(expr: StringType) -> StringExpression:
 # def treepath(self): return
 
 
-def trim(): raise NotImplemented  # TODO
+def trim() -> StringExpression: raise NotImplemented  # TODO
 
 
 # def trim_end(self): return
@@ -726,38 +741,39 @@ def trim(): raise NotImplemented  # TODO
 # def trim_start(self): return
 
 
-def url_decode(): raise NotImplemented  # TODO
+def url_decode() -> StringExpression: raise NotImplemented  # TODO
 
 
-def url_encode(): raise NotImplemented  # TODO
+def url_encode() -> StringExpression: raise NotImplemented  # TODO
 
 
-def weekofyear(): raise NotImplemented  # TODO
+def weekofyear() -> NumberExpression: raise NotImplemented  # TODO
 
 
 # def welch_test(self): return
 
 
-def zip(): raise NotImplemented  # TODO
+def zip() -> DynamicExpression: raise NotImplemented  # TODO
 
 
 # ----------------------------------------------------
 # aggregative functions
 # -----------------------------------------------------
 
-def any(*args: ExpressionType) -> AggregationExpression:
+# noinspection PyShadowingBuiltins
+def any(*args: ExpressionType) -> AnyAggregationExpression:
     res = 'any({})'.format(', '.join([arg.kql for arg in args]))
-    return AggregationExpression(KQL(res))
+    return AnyAggregationExpression(KQL(res))
 
 
-def arg_max(*args: ExpressionType) -> AggregationExpression:
+def arg_max(*args: ExpressionType) -> AnyAggregationExpression:
     res = 'arg_max({})'.format(', '.join([arg.kql for arg in args]))
-    return AggregationExpression(KQL(res))
+    return AnyAggregationExpression(KQL(res))
 
 
-def arg_min(*args: ExpressionType) -> AggregationExpression:
+def arg_min(*args: ExpressionType) -> AnyAggregationExpression:
     res = 'arg_min({})'.format(', '.join([arg.kql for arg in args]))
-    return AggregationExpression(KQL(res))
+    return AnyAggregationExpression(KQL(res))
 
 
 def avg(expr: ExpressionType) -> NumberAggregationExpression:
@@ -822,42 +838,41 @@ def make_set(expr: ExpressionType, max_size: NumberType = None) -> ArrayAggregat
 
 
 def max(expr: ExpressionType) -> AggregationExpression:
-    return AggregationExpression(KQL('max({})'.format(to_kql(expr))))
+    return AnyAggregationExpression(KQL('max({})'.format(to_kql(expr))))
 
 
 def min(expr: ExpressionType) -> AggregationExpression:
-    return AggregationExpression(KQL('min({})'.format(to_kql(expr))))
+    return AnyAggregationExpression(KQL('min({})'.format(to_kql(expr))))
 
 
 def percentile(expr: ExpressionType, per: NumberType) -> AggregationExpression:
     res = 'percentiles({}, {})'.format(expr, to_kql(per))
-    return AggregationExpression(KQL(res))
+    return AnyAggregationExpression(KQL(res))
 
 
 def percentiles(expr: ExpressionType, *pers: NumberType) -> AggregationExpression:
-    res = 'percentiles({}, {})'.format(expr.kql,
-                                       ', '.join([str(to_kql(per)) for per in pers]))
-    return AggregationExpression(KQL(res))
+    res = 'percentiles({}, {})'.format(expr.kql, ', '.join([str(to_kql(per)) for per in pers]))
+    return AnyAggregationExpression(KQL(res))
 
 
 def stdev(expr: ExpressionType) -> AggregationExpression:
-    return AggregationExpression(KQL('stdev({})'.format(to_kql(expr))))
+    return AnyAggregationExpression(KQL('stdev({})'.format(to_kql(expr))))
 
 
 def stdevif(expr: ExpressionType, predicate: BooleanType) -> AggregationExpression:
-    return AggregationExpression(KQL('stdevif({}, {})'.format(to_kql(expr), to_kql(predicate))))
+    return AnyAggregationExpression(KQL('stdevif({}, {})'.format(to_kql(expr), to_kql(predicate))))
 
 
 def stdevp(expr: ExpressionType) -> AggregationExpression:
-    return AggregationExpression(KQL('stdevp({})'.format(to_kql(expr))))
+    return AnyAggregationExpression(KQL('stdevp({})'.format(to_kql(expr))))
 
 
 def sum(expr: ExpressionType) -> AggregationExpression:
-    return AggregationExpression(KQL('sum({})'.format(to_kql(expr))))
+    return AnyAggregationExpression(KQL('sum({})'.format(to_kql(expr))))
 
 
 def sumif(expr: ExpressionType, predicate: BooleanType) -> AggregationExpression:
-    return AggregationExpression(KQL('sumif({}, {})'.format(to_kql(expr), to_kql(predicate))))
+    return AnyAggregationExpression(KQL('sumif({}, {})'.format(to_kql(expr), to_kql(predicate))))
 
 
 # def tdigest(self):
@@ -869,15 +884,15 @@ def sumif(expr: ExpressionType, predicate: BooleanType) -> AggregationExpression
 
 
 def variance(expr: ExpressionType) -> AggregationExpression:
-    return AggregationExpression(KQL('variance({})'.format(to_kql(expr))))
+    return AnyAggregationExpression(KQL('variance({})'.format(to_kql(expr))))
 
 
 def varianceif(expr: ExpressionType, predicate: BooleanType) -> AggregationExpression:
-    return AggregationExpression(KQL('varianceif({}, {})'.format(to_kql(expr), to_kql(predicate))))
+    return AnyAggregationExpression(KQL('varianceif({}, {})'.format(to_kql(expr), to_kql(predicate))))
 
 
 def variancep(expr: ExpressionType) -> AggregationExpression:
-    return AggregationExpression(KQL('variancep({})'.format(to_kql(expr))))
+    return AnyAggregationExpression(KQL('variancep({})'.format(to_kql(expr))))
 
 
 # Used for mv-expand
